import bpy
from bpy.types import Context, Event, UILayout
from bpy.props import StringProperty, CollectionProperty, IntProperty
from bpy_extras.io_utils import ImportHelper

import addon_utils

import requests
from pathlib import Path
from threading import Thread
from queue import Queue, Empty

from typing import Any, Set

from .preferences import AISTUDIO_AddonPreferences
from .properties import (
    AISTUDIO_GeneratedItem,
    AISTUDIO_Properties,
    AISTUDIO_UL_EnumManager,
)
from .generator import Generator, ListGeneration

bl_info = {
    "name": "3D AI Studio",
    "author": "3D AI Studio",
    "version": (1, 0, 0),
    "blender": (4, 1, 1),
    "description": "Generate 3D models from text and images",
    "category": "Tools",
}


class Config:
    normal_name = bl_info["name"]
    caps_name = "AISTUDIO"
    lower_case_name = caps_name.lower()
    addon_global_var_name = "ai_studio"

    panel_prefix = caps_name + "_PT_"
    operator_prefixz = caps_name + "_OT_"

    path: Path

    @staticmethod
    def get_addon_name():
        addon_path = Path(__file__).resolve().parent.stem
        return addon_path


for mod in addon_utils.modules(): # type: ignore
    if mod.bl_info["name"] == bl_info["name"]:
        filepath = mod.__file__
        Config.path = Path(filepath).parent


class BaseClass:
    @staticmethod
    def get_props() -> AISTUDIO_Properties:
        return getattr(bpy.context.scene, Config.addon_global_var_name)

    @staticmethod
    def get_preferences() -> AISTUDIO_AddonPreferences:
        return bpy.context.preferences.addons[Config.get_addon_name()].preferences


class AISTUDIO_PT_viewportSidePanel(BaseClass, bpy.types.Panel):
    bl_idname = Config.panel_prefix + "viewportSidePanel"
    bl_label = Config.normal_name
    bl_space_type = "VIEW_3D"
    bl_region_type = "UI"
    bl_context = "objectmode"
    bl_category = "3D AI Studio"

    def draw(self, context):
        layout = self.layout
        props = self.get_props()

        row = layout.row()

        row.template_list(
            listtype_name="AISTUDIO_UL_EnumManager",
            list_id="AISTUDIO_UL_EnumManager",
            dataptr=props, # type: ignore
            propname="generations",
            active_dataptr=props, # type: ignore
            active_propname="active_idx",
        )

        col = row.column(align=True)

        col.operator(AISTUDIO_OT_RefreshList.bl_idname, text="", icon="RECOVER_LAST")

        col = layout.column()

        col.scale_y = 1.4

        # col.prop(props, "api_key")  # type: ignore
        col.label(text="Enter Prompt:")
        split: UILayout = col.split(factor=0.15)

        if props.image_path == "":
            split.operator(AISTUDIO_OT_SetImage.bl_idname, text="", icon="NODE_TEXTURE")
            split.prop(props, "prompt", text="", icon="OUTLINER_DATA_LIGHTPROBE")  # type: ignore
        else:
            split.operator(AISTUDIO_OT_ClearImage.bl_idname, text="", icon="PANEL_CLOSE")
            split.label(text=props.prompt, icon="NODE_TEXTURE")  # type: ignore

        if not props.is_loading:
            col.operator(AISTUDIO_OT_Generate.bl_idname, text="Generate")
        else:
            col.label(text=props.prompt)
            col.label(text=props.loading_message)  # type: ignore


class BaseOperator(BaseClass):
    @staticmethod
    def select_and_make_active(ob: bpy.types.Object):
        for ob_to_deselect in bpy.data.objects:
            if ob_to_deselect == ob:
                continue
            ob_to_deselect.select_set(False)

        bpy.context.view_layer.objects.active = ob
        ob.select_set(True)

        print(f"[ Status ] {ob.name} set to Active Object")


class AISTUDIO_OT_Generate(BaseOperator, bpy.types.Operator):
    bl_idname = Config.lower_case_name + ".generate_text"
    bl_label = Config.normal_name + " " + "Operator"
    bl_description = "Generates a 3D object from a text prompt"
    bl_options = {"REGISTER", "UNDO"}

    props: AISTUDIO_Properties
    # num_slices: bpy.props.IntProperty(name="num_slices", default=3, min=2)

    def __init__(self) -> None:
        self.t = None
        self.timer = bpy.types.Timer
        self.queue = Queue(maxsize=1)

        self.cancelled = False

        self.spinner = [
            "⣾",
            "⣽",
            "⣻",
            "⢿",
            "⡿",
            "⣟",
            "⣯",
            "⣷",
        ]
        self.spinner_index = 0
        self.status = "Generating"

    @staticmethod
    def create_collection(collection_name, parent=None) -> bpy.types.Collection:
        if collection_name in bpy.data.collections:
            return bpy.data.collections[collection_name]

        collection = bpy.data.collections.new(collection_name)

        if parent:
            bpy.data.collections[parent].children.link(collection)
            return collection
        else:
            bpy.context.scene.collection.children.link(collection)
            return collection

    def import_glb(self, path: Path):
        print(f"Importing {path.as_posix()}")
        bpy.ops.import_scene.gltf(filepath=path.as_posix())
        ob = bpy.context.active_object

        ob.name = self.props.prompt
        ob.data.name = self.props.prompt

        collection = self.create_collection("Generated")
        collection.objects.link(ob)

        self.add_to_generated_collection(path)

    def add_to_generated_collection(self, path: Path):
        props = self.props

        item = props.generations.add()
        item.prompt = props.prompt
        item.filepath = path.as_posix()

        props.prompt = ""

        props.generations.move(len(props.generations) - 1, 0)
        props.active_idx = 0

    def cancel(self, context: Context):
        wm = context.window_manager
        wm.event_timer_remove(self.timer)  # type: ignore

        self.props.is_loading = False

        self.redraw(context)

    def done(self, context: Context):
        print("DONE.")
        self.props.loading_message = ""
        self.props.prompt = ""
        self.props.image_path = ""
        self.cancel(context)

    @staticmethod
    def redraw(context: Context):
        for area in bpy.context.screen.areas:
            if area.type == "VIEW_3D":
                for region in area.regions:
                    if region.type == "UI":
                        region.tag_redraw()

    def modal(self, context: Context, event: Event) -> Set[str]:
        if event.type in {"ESC"}:
            self.cancel(context)

            return {"CANCELLED"}

        if event.type == "TIMER":
            try:
                data: dict[str, Path] | str | None = self.queue.get(timeout=0.05)
            except Empty:
                self.props.loading_message = (
                    self.spinner[self.spinner_index] + " " + self.status
                )

                self.redraw(context)

                self.spinner_index += 1
                self.spinner_index = self.spinner_index % len(self.spinner)
                return {"PASS_THROUGH"}
            else:
                if isinstance(data, dict):
                    self.import_glb(data["model"])
                    self.done(context)
                    return {"FINISHED"}
                elif isinstance(data, str):
                    self.report({"WARNING"}, data)
                    self.cancel(context)
                    return {"FINISHED"}
                else:
                    self.report({"WARNING"}, "Could not reach API")
                    self.cancel(context)
                    return {"FINISHED"}

        return {"PASS_THROUGH"}

    def execute(self, context: Context) -> Set[str]:
        image = False

        self.props = self.get_props()
        preferences = self.get_preferences()

        if preferences.api_key == "":
            self.report({"WARNING"}, "API KEY NOT SET")
            return {"CANCELLED"}

        if self.props.prompt == "":
            self.report({"WARNING"}, "PROMPT EMPTY")
            return {"CANCELLED"}

        if self.props.image_path != "":
            image = True



        def worker():
            print("Starting threaded Download")
            g = Generator(preferences.api_key, Config.path / ".cache")

            try:
                if image:
                    res = g.image_generate(image=Path(self.props.image_path))
                else:
                    res = g.text_generate(
                        title=self.props.prompt, prompt=self.props.prompt
                    )

            except requests.RequestException as e:
                print(e)
                print("Generate Error")
                self.queue.put(None)
                return
            except Exception as e:
                print(e)
                self.queue.put("GENERATION ERROR")

            if self.cancelled:
                return

            if isinstance(res, str):
                self.queue.put(res)
                return

            self.status = "Downloading"

            try:
                thumbnail_path = g.download(
                    res["data"]["model_thumbnail"], self.cancelled
                )
                model_path = g.download(res["data"]["model_url"], self.cancelled)

            except requests.RequestException as e:
                print(e)
                print("Download Error")
                self.queue.put(None)
            else:
                self.status = "Importing"
                self.queue.put({"thumbnail": thumbnail_path, "model": model_path})

        self.t = Thread(target=worker)
        self.t.start()
        self.props.is_loading = True

        wm = context.window_manager
        self.timer = wm.event_timer_add(0.1, window=context.window)
        wm.modal_handler_add(self)

        return {"RUNNING_MODAL"}


class AISTUDIO_OT_SetImage(BaseOperator, bpy.types.Operator, ImportHelper):
    bl_idname = Config.lower_case_name + ".set_image"
    bl_label = "Select Image"
    bl_description = "Generates a 3D object from an image reference"
    bl_options = {"REGISTER", "UNDO"}

    # props: AISTUDIO_Properties

    filename_ext = {
        ".rgba",
        ".jpg",
        ".pdd",
        ".exr",
        ".psb",
        ".cin",
        ".png",
        ".tif",
        ".dds",
        ".tx",
        ".jpeg",
        ".jp2",
        ".hdr",
        ".bmp",
        ".psd",
        ".tiff",
        ".j2c",
        ".sgi",
        ".rgb",
        # ".webp",
        ".tga",
        ".dpx",
    }

    filtered = ";".join(["*" + filename for filename in filename_ext])

    filter_glob: StringProperty(default=filtered, options={"HIDDEN"})  # type: ignore

    files: CollectionProperty(
        name="File Path",
        description="File path used for importing the OBJ sequence",
        type=bpy.types.OperatorFileListElement,
    )  # type: ignore

    def execute(self, context: Context) -> Set[str]:
        self.props = self.get_props()

        path = Path(self.filepath)
        print("operator path", path)

        self.props.image_path = path.as_posix()
        self.props.prompt = path.name

        return {"FINISHED"}

class AISTUDIO_OT_ClearImage(BaseOperator, bpy.types.Operator):
    bl_idname = Config.lower_case_name + ".clear_image"
    bl_label = "Clear Image"
    bl_description = "Clears Loaded Image"
    bl_options = {"REGISTER", "UNDO"}

    def execute(self, context: Context | Any) -> Set[str]:
        props = self.get_props()
        props.image_path = ""
        props.prompt = ""
        return {"FINISHED"}


class AISTUDIO_OT_RefreshList(BaseOperator, bpy.types.Operator):
    bl_idname = Config.lower_case_name + ".refresh"
    bl_label = "Refresh"
    bl_description = "Refreshes Generation List"
    bl_options = {"REGISTER", "UNDO"}

    props: AISTUDIO_Properties

    def __init__(self) -> None:
        self.t = None
        self.timer = bpy.types.Timer
        self.queue = Queue(maxsize=1)

        self.cancelled = False

        self.spinner = [
            "⣾",
            "⣽",
            "⣻",
            "⢿",
            "⡿",
            "⣟",
            "⣯",
            "⣷",
        ]
        self.spinner_index = 0

    def add_to_generated_collection(self, data: ListGeneration):
        props = self.props

        props.generations.clear()

        for generated in data["data"]:
            item: AISTUDIO_GeneratedItem = props.generations.add()

            item.id_ = generated["id"]
            item.created_at = generated["created_at"]
            item.image = generated["image"]
            item.is_remeshed = generated["is_remeshed"]
            item.model_thumbnail = generated["model_thumbnail"]
            item.model_url = generated["model_url"]
            item.title = generated["title"]
            item.prompt = generated["prompt"]
            item.format = generated["format"]
            item.task_id = generated["task_id"]

            props.generations.move(len(props.generations) - 1, props.active_idx + 1)
            props.active_idx += 1

    def cancel(self, context: Context):
        wm = context.window_manager
        wm.event_timer_remove(self.timer)  # type: ignore

        self.props.is_loading = False
        self.props.loading_message = ""

        if self.t.is_alive():
            self.t.join()

        self.redraw(context)

    def done(self, context: Context):
        print("DONE.")
        self.cancel(context)

    @staticmethod
    def redraw(context: Context):
        for area in bpy.context.screen.areas:
            if area.type == "VIEW_3D":
                for region in area.regions:
                    if region.type == "UI":
                        region.tag_redraw()

    def modal(self, context: Context, event: Event) -> Set[str]:
        if event.type in {"ESC"}:
            self.cancel(context)

            return {"CANCELLED"}

        if event.type == "TIMER":
            try:
                data: ListGeneration = self.queue.get(timeout=0.05)
            except Empty:
                self.props.loading_message = (
                    self.spinner[self.spinner_index] + " Refreshing"
                )

                self.redraw(context)

                self.spinner_index += 1
                self.spinner_index = self.spinner_index % len(self.spinner)
                return {"PASS_THROUGH"}
            else:
                if isinstance(data, dict):
                    self.add_to_generated_collection(data)
                    self.done(context)
                    return {"FINISHED"}
                elif isinstance(data, str):
                    self.report({"WARNING"}, data.capitalize())
                    self.cancel(context)
                    return {"FINISHED"}
                else:
                    self.report({"WARNING"}, "Could not reach API")
                    self.cancel(context)
                    return {"FINISHED"}

        return {"PASS_THROUGH"}

    def execute(self, context: Context) -> Set[str]:
        self.props = self.get_props()

        preferences = self.get_preferences()

        if preferences.api_key == "":
            self.report({"WARNING"}, "API KEY NOT SET")
            return {"CANCELLED"}

        wm = context.window_manager
        self.timer = wm.event_timer_add(0.1, window=context.window)
        wm.modal_handler_add(self)

        def worker():
            print("Starting Threaded Operation")

            g = Generator(preferences.api_key, Config.path / ".cache")

            try:
                data = g.list_generations()

            except requests.RequestException as e:
                print(e)
                if e.response:
                    if e.response.status_code == 401:
                        data = e.response.json()
                        self.queue.put(data["detail"])
                        return

                print("list_generation() error")
                self.queue.put(None)
            
            except Exception as e:
                print(e)
                self.queue.put(None)
            
            else:
                self.queue.put(data)

        self.t = Thread(target=worker)
        self.t.start()
        self.props.is_loading = True

        return {"RUNNING_MODAL"}


class AISTUDIO_OT_SetPrompt(BaseOperator, bpy.types.Operator):
    bl_idname = Config.lower_case_name + ".set_prompt"
    bl_label = "Set Prompt"
    bl_description = "Set Prompt"
    bl_options = {"REGISTER"}

    props: AISTUDIO_Properties

    index: IntProperty(name="index", default=0)  # type: ignore

    def execute(self, context: Context) -> Set[str]:
        props = self.get_props()

        props.prompt = props.generations[self.index].prompt
        return {"FINISHED"}


classes = [
    AISTUDIO_UL_EnumManager,
    AISTUDIO_GeneratedItem,
    AISTUDIO_Properties,
    AISTUDIO_OT_RefreshList,
    AISTUDIO_OT_ClearImage,
    AISTUDIO_OT_SetImage,
    AISTUDIO_OT_SetPrompt,
    AISTUDIO_OT_Generate,
    AISTUDIO_PT_viewportSidePanel,
    AISTUDIO_AddonPreferences,
]


def set_properties():
    setattr(
        bpy.types.Scene,
        Config.addon_global_var_name,
        bpy.props.PointerProperty(
            type=AISTUDIO_Properties,
        ),
    )


def del_properties():
    delattr(bpy.types.Scene, Config.addon_global_var_name)


def register():
    from bpy.utils import register_class

    for cls in classes:
        register_class(cls)

    set_properties()


def unregister():
    from bpy.utils import unregister_class

    for cls in reversed(classes):
        unregister_class(cls)

    del_properties()


if __name__ == "__main__":
    register()
