import requests
from typing import TypedDict
from pathlib import Path
from urllib.parse import urlparse


class Data(TypedDict):
    model_url: str
    model_thumbnail: str
    task_id: str


class GenerateResponse(TypedDict):
    code: int
    data: Data
    message: str


class GeneratedItem(TypedDict):
    id: int
    prompt: str
    image: str
    title: str
    model_url: str
    model_thumbnail: str
    is_remeshed: bool
    format: str
    is_3d: bool
    task_id: str
    is_loading: bool
    is_favorite: bool
    created_at: str


class ListGeneration(TypedDict):
    code: int
    data: list[GeneratedItem]
    message: str


class Generator:
    def __init__(self, api_key: str, cache: Path) -> None:
        self.headers = {"Authorization": f"token {api_key}"}
        self.cache = cache
        self.done = False
        self.status = "..."

    def get(
        self, url, json: dict | None = None, files: dict | None = None
    ) -> ListGeneration | GenerateResponse | str:
        print("Hitting", url)

        try:
            r = requests.post(url, json=json, files=files, headers=self.headers)
        except requests.RequestException as e:
            print(e)
            if e.response:
                print(e.response.content)

            raise e

        else:
            print("Status code", r.status_code)

            if r.ok:
                data: ListGeneration | GenerateResponse = r.json()

                if data["code"] == 100:
                    return data

                elif data['code'] >= 400 and data['code'] < 500:
                    
                    return "Get more tokens https://3daistudio.com/#pricing"
                else:
                    print("Code:", data["code"], data["message"])
                    return data["message"]

            else:
                print(r.content)
                raise requests.RequestException

    def list_generations(self) -> ListGeneration | str:
        url = "https://s.3daistudio.com/generate/my_generations/"

        try:
            data: ListGeneration = self.get(url)
        except Exception as e:
            raise e
        else:
            return data

    def text_generate(self, title: str, prompt: str):
        url = "https://s.3daistudio.com/generate/generate/"

        post = {
            "title": title,
            "prompt": prompt,
        }

        print(f"Generating: {prompt}...")

        try:
            data: GenerateResponse = self.get(url, json=post)
        except Exception as e:
            raise e
        else:
            return data

    def image_generate(self, image: Path):
        url = "https://s.3daistudio.com/generate/img_to_3d/"

        files = {"refImage": open(image, "rb")}

        try:
            data: GenerateResponse = self.get(url, files=files)

        except Exception as e:
            raise e
        else:
            return data

    def download(self, url: str, cancel: bool):
        print("Downloading", url)

        parsed_url = urlparse(url)
        print(parsed_url)

        filepath = self.cache / parsed_url.path.split("/")[-1]
        print(filepath.as_posix())
        filepath.parent.mkdir(parents=True, exist_ok=True)
        filepath.touch()

        try:
            with requests.get(url, headers=self.headers, stream=True) as r, open(
                filepath, "wb"
            ) as f:
                if cancel:
                    raise Exception("Cancelled")

                r.raise_for_status()
                download_size = int(r.headers["Content-Length"]) / 1000**2
                content_type = r.headers.get("Content-Type")

                print("Content Type", content_type)
                print("Download Size:", download_size, "mb")

                for chunk in r.iter_content():
                    f.write(chunk)

        except KeyboardInterrupt as e:
            print(e)
            filepath.unlink(missing_ok=True)
            return

        except requests.RequestException as e:
            if e.response:
                print(e.response.content)
            raise e
        else:
            print("Download Finished")
            return filepath
